#include <LowPower.h>
#include <AES.h>
#include <AES_config.h>
#include <nRF24L01.h>
#include <printf.h>
#include <RF24.h>
#include <RF24_config.h>
#include <Wire.h>
#include <stdint.h>
#include <Adafruit_SHT4x.h>
#include <WeatherStationPacket.h>


/* -------CONFIG---------*/
#define LOGDEBUG        0
#define SLEEP_LOOP_MAX  6 //8s of sleep each loop
#define SHT41_SENSOR    1  
#define VOLTAGE_RATIO   4.9746f  //4.6237; //(1100/1023)*((33+9.1)/9.1) 
/* --------------------- */


/* -------LOGDEBUG------ */
#if LOGDEBUG==1
#define LOGINFO(...) Serial.printf(__VA_ARGS__)
#else
#define LOGINFO(...)
#endif
/* --------------------- */


/* -------PINS---------- */
#define SUPERCAP_ADC_ENABLE   16
#define SUPERCAP_ADC          A1 //15
#define LED                   14
#define RF24_CSN              5
#define RF24_CE               6
/* --------------------- */


/* -------WIRELESS------ */
#define RF24_WEATHER_CH       73
/* --------------------- */


/* ---WIRELESS ADDRESS--- */
uint8_t   weatherGatewayAddress[5]  = {0, 0, 0, 0, 0}; //Set your NRF24 receiver address here
uint8_t   weatherSensorAddress[5]   = {1, 0, 0, 0, 0}; //Set your NRF24 transmiter address here (this device)
/* --------------------- */


/* -------AES KEYS------ */
uint8_t   weatherAesKey[32] = {0x6d, 0x56, 0x1c, 0x2d, 0xf0, 0xe4, 0xc5, 0x97, 0x81, 0x8a, 0x27, 0x88, 0x5d, 0xd3, 0x8d, 0x10, 0xf0, 0x37, 0x8e, 0x56, 0x51, 0x02, 0x14, 0x80, 0xc3, 0xb8, 0x78, 0x3d, 0x2f, 0x9f, 0x09 , 0x11}; 
/* --------------------- */
                   
WeatherStationPacket weatherStationPacket(SHT41_SENSOR);
AES       aes;

RF24      radio(RF24_CE, RF24_CSN, 2000000);

uint8_t sleepLoop = SLEEP_LOOP_MAX;

Adafruit_SHT4x sht41 = Adafruit_SHT4x();
sensors_event_t   temperature;
sensors_event_t   humidity;
uint16_t          supercapVoltage;

enum sensorStatus {
  SHT4x_OK,
  SHT4x_NOT_FOUND,
  SHT4x_READ_FAILED, // 2
};

uint8_t sht4xStatus = SHT4x_OK;




uint16_t getCapacitorVoltage(){
  //enable voltage divisor
  LOGINFO("enable voltage divisor\n");
  digitalWrite(SUPERCAP_ADC_ENABLE, HIGH);
  delay(10);

  float voltage = analogRead(SUPERCAP_ADC);
  voltage = voltage * VOLTAGE_RATIO; 

  //disable voltage divisor
  LOGINFO("disable voltage divisor\n");
  digitalWrite(SUPERCAP_ADC_ENABLE, LOW);  

  LOGINFO("Voltage value: %hu\n", (uint16_t)voltage);
  return (uint16_t)voltage;
}




void setup() {
  delay(500);

  pinMode(SUPERCAP_ADC_ENABLE , OUTPUT);
  digitalWrite(SUPERCAP_ADC_ENABLE, LOW);

  pinMode(LED , OUTPUT);
  digitalWrite(LED, HIGH);

  analogReference(INTERNAL1V1);

  #if LOGDEBUG==1
  Serial.begin(250000);
  printf_begin();
  #endif

  radio.begin();
  radio.setChannel(RF24_WEATHER_CH);
  radio.setDataRate(RF24_250KBPS);
  radio.setPALevel(RF24_PA_MAX);
  radio.setCRCLength(RF24_CRC_16);
  radio.setRetries(15, 3);  
  radio.setAutoAck(1);
  radio.openReadingPipe(1, weatherSensorAddress);
  radio.openWritingPipe(weatherGatewayAddress);

  #if LOGDEBUG==1
  radio.printDetails();
  #endif

  //SHT41
  LOGINFO("Init SHT4x\n");
  if (! sht41.begin()) {
    LOGINFO("Couldn't find SHT4x");
    sht4xStatus = SHT4x_NOT_FOUND;
  }
  
  LOGINFO("Found SHT4x\n");

  sht41.setPrecision(SHT4X_HIGH_PRECISION);
  sht41.setHeater(SHT4X_NO_HEATER);

  aes.set_key(weatherAesKey, 256);

  delay(500);
  digitalWrite(LED, LOW);
}



void loop() {  
  if( sleepLoop >= SLEEP_LOOP_MAX){
    supercapVoltage = getCapacitorVoltage();

    if(sht4xStatus != SHT4x_NOT_FOUND){
      if(sht41.getEvent(&humidity, &temperature)){
        sht4xStatus = SHT4x_OK;
        LOGINFO("Read OK\n");
      } else {
        sht4xStatus = SHT4x_READ_FAILED;
        LOGINFO("Read failed\n");
      }
  
      #if LOGDEBUG==1
      Serial.print("SuperCap voltage: "); Serial.print(supercapVoltage); Serial.println(" mV");
      Serial.print("Temperature: "); Serial.print(temperature.temperature); Serial.println(" degrees C");
      Serial.print("Humidity: "); Serial.print(humidity.relative_humidity); Serial.println("% rH");
      #endif   

      //use heater next time if humidity is high (consended water on sensor ?)
      if(humidity.relative_humidity > 99){
        sht41.setHeater(SHT4X_LOW_HEATER_1S);
      }
      else {
        sht41.setHeater(SHT4X_NO_HEATER);
      }      
    }

    //Send to weather station
    LOGINFO("Send to weather station\n");
    weatherStationPacket.setType(0);
    weatherStationPacket.setStatus((uint8_t)sht4xStatus);

    weatherStationPacket.write(4, &temperature.temperature, 4);
    weatherStationPacket.write(8, &humidity.relative_humidity, 4);
    weatherStationPacket.write(12, &supercapVoltage, 2);

    weatherStationPacket.updateCRC();

  #if LOGDEBUG==1
    weatherStationPacket.printPlainBuffer();
  #endif
    

    aes.set_key(weatherAesKey, 256);
    aes.encrypt(weatherStationPacket.getPlainBuffer(),     weatherStationPacket.getCipherBuffer());
    aes.encrypt(weatherStationPacket.getPlainBuffer()+16,     weatherStationPacket.getCipherBuffer()+16);

    radio.setChannel(RF24_WEATHER_CH);
    radio.openReadingPipe(1, weatherSensorAddress);
    radio.openWritingPipe(weatherGatewayAddress);
    
    radio.powerUp();
    delay(10); 
    radio.write(weatherStationPacket.getCipherBuffer(), 32);
    radio.powerDown();
   
    
    sleepLoop = 0;
  } else {
    #if LOGDEBUG==1
    LOGINFO(".");
    delay(10);  
    #endif    
  
    ++sleepLoop;
  }


  LowPower.powerDown(SLEEP_8S, ADC_OFF, BOD_ON);
}
